/* 
 * File:   rsuk.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 29. května 2019, 10:52
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <glib.h>

#include "gdgnet.h"
#include "gdgelement.h"
#include "gdgpart.h"
#include "main.h"

#include "rsuk.h"


int rsuk_get_loc_x ( int rs_loc_x ) {
    if ( rs_loc_x <= 50 ) return ( rs_loc_x + 1 );
    if ( rs_loc_x <= 101 ) return ( rs_loc_x + 3 );
    return ( rs_loc_x + 5 );
}


int rsuk_get_loc_y ( int rs_loc_y ) {
    return ( rs_loc_y + 1 );
}


int rsuk_get_rs_loc_x ( int loc_x ) {
    if ( loc_x <= 51 ) return ( loc_x - 1 );
    if ( loc_x <= 104 ) return ( loc_x - 3 );
    return ( loc_x - 5 );
}


int rsuk_get_rs_loc_y ( int loc_y ) {
    return ( loc_y - 1 );
}


int rsuk_get_rs_block_from_rs_loc ( int rs_loc_x, int rs_loc_y ) {
    return ( rs_loc_x + ( rs_loc_y * RSUK_LOC_WIDTH ) );
}


int rsuk_get_rs_block_from_loc ( int loc_x, int loc_y ) {
    int rs_loc_x = rsuk_get_rs_loc_x ( loc_x );
    int rs_loc_y = rsuk_get_rs_loc_y ( loc_y );
    return rsuk_get_rs_block_from_rs_loc ( rs_loc_x, rs_loc_y );
}


void rsuk_get_loc_from_rs_block ( int block, int *loc_x, int *loc_y ) {
    *loc_x = rsuk_get_loc_x ( block % RSUK_LOC_WIDTH );
    *loc_y = rsuk_get_loc_y ( block / RSUK_LOC_WIDTH );
}


st_GDGPARTSALL* rsuk_gdgparts_init ( st_GDGELEMENTSALL *elmall, const char *filepath ) {
    FILE *fp = fopen ( filepath, "r" );
    if ( !fp ) {
        GString *str = g_string_new ( "" );
        g_string_append_printf ( str, "%s(%d) - ERROR: Can't open input file '%s'\n", __func__, __LINE__, filepath );
        main_error_exit ( g_string_free ( str, FALSE ) );
        return NULL;
    };

    st_GDGPARTSALL *partall = gdgparts_new ( elmall, GDGPARTSALL_TYPE_MASTER );

    char buf[1024];

    GRegex *regex = g_regex_new ( "^(\\d+);(\\d+);B(\\d+);([IO]);([\\w|/|_|\\-]+);(F\\d+);.*", 0, 0, NULL );

    while ( fgets ( buf, sizeof ( buf ), fp ) ) {
        GMatchInfo *match_info;

        g_regex_match ( regex, buf, 0, &match_info );
        if ( g_match_info_matches ( match_info ) ) {
            gchar *element_name = g_match_info_fetch ( match_info, 6 );
            gchar *block_txt = g_match_info_fetch ( match_info, 3 );

            int block = atoi ( block_txt );

            if ( !( ( block == 9999 ) || ( 0 == strcmp ( element_name, "F000" ) ) || ( 0 == strcmp ( element_name, "F961" ) ) ) ) {

                int loc_x = 0;
                int loc_y = 0;
                rsuk_get_loc_from_rs_block ( block, &loc_x, &loc_y );

                if ( 0 == strcmp ( element_name, "F962" ) ) {
                    element_name[3]--;
                };

                st_GDGELEMENT *elm = gdgelements_search_element_by_name ( elmall, element_name );
                if ( !elm ) {
                    fclose ( fp );
                    return NULL;
                };

                st_GDGPART *part = gdgparts_get_part_by_location ( partall, loc_x, loc_y, FALSE );
                if ( ( !part ) || ( part->element != elm ) ) {
                    gdgparts_add_part ( partall, elm, loc_x, loc_y );
                };

                if ( !gdgparts_reindex_part_locations ( partall ) ) return NULL;

            };

            g_free ( element_name );
            g_free ( block_txt );
            g_match_info_next ( match_info, NULL );
        }
        g_match_info_free ( match_info );
    };

    g_regex_unref ( regex );

    fclose ( fp );

    fprintf ( stderr, "%s() - initialised %d parts\n", __func__, partall->count_part );

    //if ( !gdgparts_reindex_part_locations ( partall ) ) return NULL;

    return partall;
}


st_GDGNETSALL* rsuk_gdgnets_init ( st_GDGELEMENTSALL *elmall, const char * filepath ) {


    typedef struct st_LOC {
        int x;
        int y;
    } st_LOC;

    int count_F961 = 0;
    st_LOC *used_F961 = g_new0 ( st_LOC, 1 );

    FILE *fp = fopen ( filepath, "r" );
    if ( !fp ) {
        GString *str = g_string_new ( "" );
        g_string_append_printf ( str, "%s(%d) - ERROR: Can't open input file '%s'\n", __func__, __LINE__, filepath );
        main_error_exit ( g_string_free ( str, FALSE ) );
        return NULL;
    };

    st_GDGNETSALL *netall = gdgnets_new ( GDGNETSALL_TYPE_MASTER );

    char buf[1024];
    int row = 1;

    GRegex *regex = g_regex_new ( "^(\\d+);(\\d+);B(\\d+);([IO]);([\\w|/|_|\\-]+);((F\\d+){0,1});.*", 0, 0, NULL );

    while ( fgets ( buf, sizeof ( buf ), fp ) ) {
        GMatchInfo *match_info;
        if ( g_regex_match ( regex, buf, 0, &match_info ) ) {

            gchar *net_name = g_match_info_fetch ( match_info, 2 );
            gchar *block_txt = g_match_info_fetch ( match_info, 3 );
            gchar *dirrection = g_match_info_fetch ( match_info, 4 );
            gchar *pin_name = g_match_info_fetch ( match_info, 5 );
            gchar *element_name = g_match_info_fetch ( match_info, 6 );

            st_GDGNET *net = gdgnets_get_net_by_name ( netall, net_name );
            if ( !net ) {
                net = gdgnet_create ( net_name );
                gdgnets_add_net ( netall, net );
            };

            int block = atoi ( block_txt );

            st_GDGELEMENT *elm = NULL;
            int loc_x = 0;
            int loc_y = 0;
            int loc_z = 0;
            en_GDGNET_REALPIN_TYPE rptype;

            gboolean ignored_pin = FALSE;

            if ( ( block == 9999 ) || ( 0 == strncmp ( pin_name, "VCC", 3 ) ) || ( 0 == strncmp ( pin_name, "GND", 3 ) ) ) {

                loc_x = loc_y = loc_z = 9999;

                if ( block == 9999 ) {
                    // skutecny I/O pin?

                    if ( 0 == strcmp ( dirrection, "O" ) ) {
                        rptype = GDGNET_REALPIN_TYPE_IN;
                    } else {
                        rptype = GDGNET_REALPIN_TYPE_OUT;
                    };

                    if ( pin_name[0] == '/' ) {
                        pin_name[0] = 'n';
                    };

                    loc_x = loc_y = loc_z = atoi ( net->name );

                } else if ( pin_name[0] == 'V' ) {
                    // VCC
                    ignored_pin = TRUE;
                    net->is_constant = TRUE;
                    net->c_value = TRUE;
                    loc_x = loc_y = 1;
                    loc_z = 0;
                } else {
                    // GND
                    ignored_pin = TRUE;
                    net->is_constant = TRUE;
                    net->c_value = FALSE;
                    loc_x = loc_y = 2;
                    loc_z = 0;
                };

                //printf ( "%s\n", pin_name );
            } else {

                rsuk_get_loc_from_rs_block ( block, &loc_x, &loc_y );

                if ( 0 == strcmp ( element_name, "F961" ) ) {
                    loc_y++;
                } else if ( 0 == strcmp ( element_name, "F962" ) ) {
                    element_name[3]--;
                    gboolean found = FALSE;
                    int i;
                    for ( i = 0; i < count_F961; i++ ) {
                        if ( ( used_F961[i].x == loc_x ) && ( used_F961[i].y == loc_y ) ) {
                            found = TRUE;
                            break;
                        };
                    };
                    if ( !found ) {
                        int j = count_F961++;
                        used_F961 = g_renew ( st_LOC, used_F961, count_F961 );
                        used_F961[j].x = loc_x;
                        used_F961[j].y = loc_y;
                    };
                };


                elm = gdgelements_search_element_by_name ( elmall, element_name );
                if ( !elm ) {
                    fprintf ( stderr, "%s(%d) - ERROR: parsing row! (row: %d)\n", __func__, __LINE__, row );
                    fprintf ( stderr, "%s", buf );
                    return NULL;
                };

#if 1
                if ( ( 0 == strcmp ( pin_name, "/RST" ) ) || ( 0 == strcmp ( pin_name, "/RESET" ) ) ) {
                    strncpy ( pin_name, "RST", strlen ( pin_name ) );
                } else if ( 0 == strcmp ( pin_name, "/SET" ) ) {
                    strncpy ( pin_name, "SET", strlen ( pin_name ) );
                } else if ( 0 == strcmp ( pin_name, "CARRY_IN" ) ) {
                    strncpy ( pin_name, "CI", strlen ( pin_name ) );
                } else if ( 0 == strcmp ( pin_name, "/CARRY_OUT" ) ) {
                    strncpy ( pin_name, "CO", strlen ( pin_name ) );
                } else if ( 0 == strcmp ( pin_name, "ENABLE" ) ) {
                    strncpy ( pin_name, "EN", strlen ( pin_name ) );
                } else if ( 0 == strncmp ( pin_name, "/CARRY", 5 ) ) {
                    char c = pin_name[6];
                    strncpy ( pin_name, "CO1", strlen ( pin_name ) );
                    pin_name[2] = c;
                } else if ( 0 == strcmp ( pin_name, "A" ) ) {
                    pin_name = g_renew ( char, pin_name, 4 );
                    strncpy ( pin_name, "INA", 4 );
                } else if ( 0 == strcmp ( pin_name, "B" ) ) {
                    pin_name = g_renew ( char, pin_name, 4 );
                    strncpy ( pin_name, "INB", 4 );
                    //} else if ( 0 == strncmp ( pin_name, "VCC", 3 ) ) {
                    //ignored_pin = TRUE;
                    //printf ( "%s - %s\n", pin_name, element_name );
                    //printf ( "%s", buf );
                };
#endif
                //printf ( "%s - %s - %s\n", element_name, dirrection, pin_name );
                //printf ( "%s - %s\n", dirrection, pin_name );

                if ( 0 == strcmp ( dirrection, "O" ) ) {
                    rptype = GDGNET_REALPIN_TYPE_OUT;
                } else {
                    rptype = GDGNET_REALPIN_TYPE_IN;
                };
            };

            if ( !ignored_pin ) {
                st_GDGNET_PIN *pin = gdgnet_pin_create ( elm, net, pin_name, rptype, loc_x, loc_y, loc_z );
                gdgnet_add_pin ( net, pin );
                if ( elm != NULL ) {
                    if ( ( 0 == strncmp ( pin_name, "OUT", 3 ) ) || ( 0 == strncmp ( pin_name, "CO", 2 ) ) ) {
                        net->count_output++;
                    };
                } else {
                    net->count_externals++;
                };
            }

            g_free ( net_name );
            g_free ( block_txt );
            g_free ( dirrection );
            g_free ( pin_name );
            g_free ( element_name );

        } else {
            fprintf ( stderr, "%s(%d) - ERROR: parsing row! (row: %d)\n", __func__, __LINE__, row );
            fprintf ( stderr, "%s", buf );
            return NULL;
        };

        g_match_info_free ( match_info );
        row++;
    };

    g_regex_unref ( regex );

    fclose ( fp );

    fprintf ( stderr, "%s() - parsed %d input rows\n", __func__, row );
    fprintf ( stderr, "%s() - initialised %d nets\n", __func__, netall->count_net );

    // bugfix pro otevrene outputy CO2 u F961
    st_GDGELEMENT *elm = gdgelements_search_element_by_name ( elmall, "F961" );
    int i;
    for ( i = 0; i < count_F961; i++ ) {
        GString *str = g_string_new ( "" );
        g_string_append_printf ( str, "nc_%d", i );
        st_GDGNET *net = gdgnet_create ( str->str );
        g_string_free ( str, TRUE );
        st_GDGNET_PIN *pin = gdgnet_pin_create ( elm, net, "CO2", GDGNET_REALPIN_TYPE_OUT, used_F961[i].x, used_F961[i].y, 0 );
        gdgnet_add_pin ( net, pin );
        net->count_output++;
        //printf ( "ADD %s_%d_%d CO2 - %s\n", elm->name, pin->loc_x, pin->loc_y, net->name );
        gdgnets_add_net ( netall, net );
    };

    g_free ( used_F961 );

    if ( !gdgnet_check ( netall ) ) return NULL;


    // fixneme nazvy konstant
    for ( i = 0; i < netall->count_net; i++ ) {
        st_GDGNET *net = netall->net[i];
        if ( net->is_constant ) {
            GString *str = g_string_new ( "" );
            g_string_append_printf ( str, "const_%s", net->name );
            g_free ( net->name );
            net->name = g_string_free ( str, FALSE );
            //printf("fixed net name: %s\n", net->name);
        };
    };

    return netall;

}


/*
 * 
 * Compare
 * 
 */
void rsuk_compare_parts ( st_GDGPARTSALL *spice_partall, st_GDGPARTSALL *rsuk_partall ) {

    fprintf ( stderr, "COMPARE - PARTS\n" );

    int count_errors = 0;
    int i;

    for ( i = 0; i < rsuk_partall->count_part; i++ ) {
        st_GDGPART *rs_part = rsuk_partall->part[i];
        st_GDGPART *spice_part = gdgparts_get_part_by_location ( spice_partall, rs_part->loc_x, rs_part->loc_y, FALSE );
        char *part_name = gdgpart_create_name ( rs_part );
        if ( !spice_part ) {
            printf ( "Part not found in SPICE '%s' (B%04d)\n", part_name, rsuk_get_rs_block_from_loc ( rs_part->loc_x, rs_part->loc_y ) );
            count_errors++;
        } else if ( 0 != strcmp ( spice_part->element->name, rs_part->element->name ) ) {
            printf ( "Part '%s' (B%04d) is not same in spice: %s\n", part_name, rsuk_get_rs_block_from_loc ( rs_part->loc_x, rs_part->loc_y ), spice_part->element->name );
            count_errors++;
        };
        g_free ( part_name );
    };

    for ( i = 0; i < spice_partall->count_part; i++ ) {
        st_GDGPART *spice_part = spice_partall->part[i];
        st_GDGPART *rs_part = gdgparts_get_part_by_location ( rsuk_partall, spice_part->loc_x, spice_part->loc_y, FALSE );
        char *part_name = gdgpart_create_name ( spice_part );
        if ( !rs_part ) {
            printf ( "Part not found in RSUK '%s' (B%04d)\n", part_name, rsuk_get_rs_block_from_loc ( spice_part->loc_x, spice_part->loc_y ) );
            count_errors++;
        };
        g_free ( part_name );
    };

    printf ( "COMPARE PARTS DONE WIDTH: %d ERROR(S)\n", count_errors );
}


static char* rsuk_get_net_source_txt ( st_GDGPARTSALL *partall, st_GDGNET *net ) {
    GString *str = g_string_new ( "" );
    st_GDGNET_PIN *pin = gdgnet_get_net_source_pin ( net );
    if ( !pin ) return g_string_free ( str, FALSE );
    if ( pin->element ) {
        st_GDGPART *part = gdgparts_get_part_by_location ( partall, pin->loc_x, pin->loc_y, TRUE );
        char *part_name = gdgpart_create_name ( part );
        g_string_append_printf ( str, "%s @ %s", part_name, pin->part_pin_name );
        g_free ( part_name );
        return g_string_free ( str, FALSE );
    };
    g_string_append_printf ( str, "PIN @ %s", pin->part_pin_name );
    return g_string_free ( str, FALSE );
}


void rsuk_compare_netlist ( st_GDGPARTSALL *src_partall, st_GDGPARTSALL *dst_partall, gboolean check_all_ppins ) {

    fprintf ( stderr, "COMPARE - NETLIST\n" );

    int count_errors = 0;
    int count_OK = 0;
    int i;

    int count_checked_nets = 0;
    st_GDGNET **checked_nets = g_new0 ( st_GDGNET*, 1 );

    for ( i = 0; i < src_partall->count_part; i++ ) {
        st_GDGPART *src_part = src_partall->part[i];
        st_GDGPART *dst_part = gdgparts_get_part_by_location ( dst_partall, src_part->loc_x, src_part->loc_y, FALSE );

        char *part_name = gdgpart_create_name ( src_part );
        GString *str = g_string_new ( "" );
        g_string_append_printf ( str, "'x%s' (B%04d)", part_name, rsuk_get_rs_block_from_loc ( src_part->loc_x, src_part->loc_y ) );
        g_free ( part_name );
        part_name = g_string_free ( str, FALSE );

        if ( !dst_part ) {
            printf ( "DST Part not found %s\n", part_name );
            count_errors++;
        } else if ( 0 != strcmp ( src_part->element->name, dst_part->element->name ) ) {
            printf ( "Part %s is not same: %s\n", part_name, dst_part->element->name );
            count_errors++;
        } else {
            st_GDGELEMENT *elm = src_part->element;
            int j;
            for ( j = 0; j < elm->count_port; j++ ) {
                st_GDGNET *src_net = src_part->netpin[j];

                int k;
                gboolean is_checked = FALSE;
                if ( !check_all_ppins ) {
                    for ( k = 0; k < count_checked_nets; k++ ) {
                        if ( checked_nets[k] == src_net ) {
                            is_checked = TRUE;
                            break;
                        };
                    };
                }

                if ( !is_checked ) {

                    if ( !check_all_ppins ) {
                        k = count_checked_nets++;
                        checked_nets = g_renew ( st_GDGNET*, checked_nets, count_checked_nets );
                        checked_nets[k] = src_net;
                    };

                    st_GDGNET *dst_net = dst_part->netpin[j];

                    st_GDGNET_PIN *src_ppin = gdgnet_get_net_source_pin ( src_net );
                    st_GDGNET_PIN *dst_ppin = gdgnet_get_net_source_pin ( dst_net );

                    char *src_ppin_txt = rsuk_get_net_source_txt ( src_partall, src_net );
                    char *dst_ppin_txt = rsuk_get_net_source_txt ( dst_partall, dst_net );


                    if ( ( src_net->is_constant != dst_net->is_constant )
                         || ( ( src_net->is_constant ) && ( src_net->c_value != dst_net->c_value ) )
                         || ( ( src_ppin ) && ( dst_ppin ) && ( src_ppin->element ) && ( dst_ppin->element ) && ( 0 != strcmp ( src_ppin_txt, dst_ppin_txt ) ) )
                         || ( ( ( src_ppin ) && ( !dst_ppin ) ) || ( ( !src_ppin ) && ( dst_ppin ) ) )
                         || ( ( ( ( src_ppin ) && ( src_ppin->loc_z ) ) || ( ( dst_ppin ) && ( dst_ppin->loc_z ) ) ) && ( 0 != strcmp ( src_ppin_txt, dst_ppin_txt ) ) ) ) {
                        count_errors++;
                        printf ( "Different power source on %s @ %s:\n", part_name, src_part->element->port[j]->name );
                        // TODO: zobrazit Z
                        printf ( "\tNET %s => %s\n", src_net->name, ( !src_net->is_constant ) ? src_ppin_txt : ( ( src_net->c_value ) ? "const_VCC" : "const_GND" ) );
                        printf ( "\tNET %s => %s\n\n", dst_net->name, ( !dst_net->is_constant ) ? dst_ppin_txt : ( ( dst_net->c_value ) ? "const_VCC" : "const_GND" ) );
                    } else if ( ( !src_net->is_constant ) ) {
                        if ( 0 != strcmp ( src_ppin_txt, dst_ppin_txt ) ) {
                            count_errors++;
                            printf ( "Different power source on %s @ %s:\n", part_name, src_part->element->port[j]->name );
                            // TODO: zobrazit Z
                            printf ( "\tNET %s => %s\n", src_net->name, ( !src_net->is_constant ) ? src_ppin_txt : ( ( src_net->c_value ) ? "const_VCC" : "const_GND" ) );
                            printf ( "\tNET %s => %s\n\n", dst_net->name, ( !dst_net->is_constant ) ? dst_ppin_txt : ( ( dst_net->c_value ) ? "const_VCC" : "const_GND" ) );
                        } else {
                            count_OK++;
                        };
                    } else {
                        count_OK++;
                    };

                    g_free ( src_ppin_txt );
                    g_free ( dst_ppin_txt );


                    /*
                       if ( src_net->count_pin != dst_net->count_pin ) {
                           count_errors++;
                           printf ( "diff: %d, %d\n", src_net->count_pin, dst_net->count_pin );
                       };
                     */
                };
            };
        };
        g_free ( part_name );
    };

    g_free ( checked_nets );

    printf ( "COMPARE NETLIST DONE WIDTH: %d ERROR(S), %d OK\n", count_errors, count_OK );
}
